/* BrPred.H
 * Private/Shared Branch Predictor for multithreaded processors
 * predictor that uses xor of PC and Global Branch History Reg
 * Parameters: predictor size, BHR size
 */

#ifndef AINT_BRPRED_H
#define AINT_BRPRED_H

#include <sys/types.h>

int UpdateTaken[] = {1, 2, 3, 3};
int UpdateNotTaken[] = {0, 0, 1, 2};

class BrPred {
public:
  int Height;
  ushort BHR_size; // Max of 31 - 2G entries!
  int BHR_mask;

  ushort IndexBits;
  int BHR;
  uchar *Counters;

  ulong Refs;
  ulong Mispredictions;

public:
  BrPred (int height, ushort bhr_size)
  {
    int i;
    Height=height;
    BHR_size = bhr_size;
    BHR_mask = (BHR_size<<1)-1;
    IndexBits = Log2(Height);
    BHR = 0x0;
    Refs = 0;
    Mispredictions = 0;
    Counters = new uchar[Height];
    // Initialize to weakly not taken
    for (i=0; i<Height; i++) {
      Counters[i] = 1;
    }
  }

  int Predict(ulong PC)
  {
    int index = (PC>>2) & (Height-1);
    index = (index ^ BHR);
    Refs++;
    return (Counters[index]>>1);
  }

  void Update(ulong PC, int Taken)
  {
    int index = ( (PC>>2) & (Height-1)) ^ BHR;
    Mispredictions += ((Counters[index]>>1) == Taken);
    if (Taken) {
      Counters[index] = UpdateTaken[Counters[index]];
    } else {
      Counters[index] = UpdateNotTaken[Counters[index]];
    }
    BHR = ((BHR<<1)+Taken)&BHR_mask;
  }
};

#endif
