

// simple.C: a simple non-pipelined program-order processor simulation

#include <iostream.h>

#include <aint.h>

#include <ICache.H>

// The catch-all include file that declares all relevant AINT types, consts,
// variables, and objects

aint_addr_t PC;    // address to fetch from
path_ptr ThePath;  // Main path in the main thread in the target binary

ICache icache(256, 32, 1);

// Aint calls this function before opening any object files - use this
// to initialize back-end variables
int
sim_init(int argc, char **argv, char **envp)
{
  /* AINT uses the getopt() function to parse the argument string, and
   * expects optind to have the correct value (first object file name)
   * upon return from this function. We must therefore use getopt() to
   * process arguments meant for us (delimited by "--").
   */

  extern int optind;    // AINT uses the getopt() function to keep track of args

  // For this example, we use no arguments...

  // Return optind so that AINT finds the name of the target binary...
  return optind;
}


// Aint calls this function if the command line has an error
void
sim_usage(int argc, char **argv)
{
  cerr << endl << "Usage: <simname> [AINT options] executable" << endl;
  cerr << "This simulator takes no arguments, and accepts one executable" << endl;
}


// AINT calls this function when it creates the main thread in the process
// corresponding to the target binary
void
sim_thread_begin(int tid, aint_addr_t pc)
{
  thread_ptr pthread = Threads[tid];    // Get thread pointer from AINT's table
  ThePath = AINT_create_path(tid, pc);    // Ask AINT to create main path

  informative("sim_thread_begin: began thread/path" << pthread->tid
      << " " << (void *) pthread->path_tree->pathid << endl);

  PC = pc;    // Starting address for thread
}


// The main execution loop - fetch, execute, commit each inst one at a time
void
sim_execute()
{
  inflight_inst_ptr ifi;

  informative("sim_execute: commencing execution at " << (void *) PC << endl);

  while(PC) {    // we get a zero PC in *ifi when program executes _exit()

    ifi = AINT_fetch(ThePath, PC);    // fetch next instruction
    informative("Fetch: " << (void *) PC << endl);

    icache.Fetch(PC, 0);

    AINT_issue(ifi);    // ask AINT to issue the fetched instruction
    informative("Issue: " << (void *) PC << endl);

    if (ifi->iflags & E_READ) {    // this is a load inst
      AINT_do_read(ifi);
    } else if (ifi->iflags & E_WRITE) {    // this is a store inst
      AINT_do_write(ifi);
    }

    PC = ifi->nextPC;    // AINT updates ifi->nextPC upon executing inst

    AINT_commit(ifi);    // Done
  }

}  // sim_execute()


void
sim_done()
{
  icache.PrintStats();
}

