/* storebuf.h
 * the microarchitectural store buffer
 */

#ifndef STOREBUF_H
#define STOREBUF_H

const int StoreBuffer_Size = 32;

class stb_entry {
public:
  // Insert stb_entry into existing list after _stb_entry
  stb_entry(stb_entry *_stb_entry, inflight_inst_ptr _ifi) {
    ifi = _ifi; fetchnum = ifi->fetchnum;
    next = _stb_entry->next; _stb_entry->next = this;
    next->prev = this; prev = _stb_entry;
  }

  // Initialize as qhead
  stb_entry() {
    ifi = NULL; fetchnum = -1ul;
    prev = next = this;
  }

  ~stb_entry() {
    prev->next = next; next->prev = prev;
  }

  stb_entry *next;
  stb_entry *prev;

  inflight_inst_ptr ifi;
  ulong fetchnum;
};

class stb_addr {
public:
  // Initialize as a qhead
  stb_addr() { next = prev = this; updates = new stb_entry; addr = (aint_addr_t) NULL; }

  // Insert after _stb_addr in an existing doubly-linked list
  stb_addr(class stb_addr *_stb_addr, aint_addr_t _addr) {
    next = _stb_addr->next; _stb_addr->next = this;
    prev = _stb_addr; next->prev = this;
    updates = new stb_entry; addr = _addr;
  }

  ~stb_addr() {
    delete updates;
    prev->next = next; next->prev = prev;
  }

  class stb_addr *next;
  class stb_addr *prev;

  aint_addr_t addr;
  stb_entry *updates;
};

class Store_Buffer {
public:
  Store_Buffer() { StoreQ = new stb_addr(); }
  
  void Update(inflight_inst_ptr ifi);
  void Remove(inflight_inst_ptr ifi);
  void Lookup(inflight_inst_ptr ifi, ulong &bitmap);

protected:
  class stb_addr *StoreQ;
};

#define QSEARCH(_Qhead, _Addr, _p) {for(_p=_Qhead->next; _p!=_Qhead; _p=_p->next){ if (_p->addr==_Addr) break; } }

#endif
