/* exec.C
 * contains all functions pertaining to execution of instructions
 */

// include files go here
#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/sysinfo.h>
#include <setjmp.h>
#include <machine/fpu.h>
#include <machine/hal_sysinfo.h>
#include <assert.h>
#include <signal.h>

#include <aint_types.h>
#include <aint_consts.h>
#include <aint.h>
#include <globals.h>
#include <protos.h>
#include <opcodes.h>

void execute(void)
{
  // For now, simply call the pmint_execute function in the backend
  sim_execute();

}

void aint_done(void)
{
  sim_done();
  exit(0);
}



Physical_RegNum
thread::ReMap(path_ptr ppath, Logical_RegNum reg)
{
  Physical_RegNum pr;

  if (reg == ZERO_REGISTER) {
    Set_InUse(ZERO_REGISTER);
    return ZERO_REGISTER;
  }

  if (FreeRegStackTop < 0) {
    fatal("ReMap: Cannot remap register in thread " << tid << endl);
  } else {
    pr = FreeRegStack[FreeRegStackTop];
    FreeRegStackTop--;
    assert(!(Is_InUse(pr)));
    ppath->RegMap[reg] = pr;
    Set_InUse(pr);
    Reg[pr].Int64 = 0;
    return pr;
  }
}

// #define USE_OBJHOOK
#ifdef USE_OBJHOOK
int ObjHookProc()
{
  cerr << "Reached ObjHookProc!" << endl;
  return 0;
}
#endif

// This function is called by the backend simulator to cause
// Aint to fetch an instruction for a path from a given PC
inflight_inst_ptr
AINT_fetch(path_ptr ppath, aint_addr_t Addr)
{
  inflight_inst_ptr ifi;


  // For debug only. #define out to optimize.
#ifdef USE_OBJHOOK
  if (Addr == ObjHook) {
    ObjHookProc();
  }
#endif

  if (ppath->instwin->is_full()) return (inflight_inst_ptr) NULL;

  icode_ptr picode = ppath->pthread->pprocess->addr2iphys(Addr, NULL);
  if (picode == (icode_ptr) NULL) {
    // iTLB fault
    return (inflight_inst_ptr) NULL;
  } else {
    ifi = new inflight_inst (ppath, picode);

    // Register renaming
    ifi->args[RA] = ppath->RegMap[picode->args[RA]];
    ifi->args[RB] = ppath->RegMap[picode->args[RB]];
    ifi->args[RC] = ppath->RegMap[picode->args[RC]];

    // #define DEBUG_REGMAP
#ifdef DEBUG_REGMAP
    assert(ppath->pthread->Is_InUse(ifi->args[RA]));
    assert(ppath->pthread->Is_InUse(ifi->args[RB]));
    assert(ppath->pthread->Is_InUse(ifi->args[RC]));
#endif

    // #define DEBUG_W1
#ifdef DEBUG_W1
    for (int i=0; i<2*N_Logical_Regs-1; i++) {
      if (!ppath->pthread->Is_InUse(ppath->RegMap[i])) {
	cerr << "mapped register not in use?" << endl;
      }
    }
#endif

    ifi->OldDest_Phys = ZERO_REGISTER;
    ifi->NewDest_Phys = ZERO_REGISTER;
    if (picode->dest < MaxArgs) {
      if (picode->args[picode->dest] != ZERO_REGISTER) {
	/* This instruction writes a register. Rename that register */
	ifi->OldDest_Phys = ifi->args[picode->dest];
	ifi->args[picode->dest] = ppath->pthread->ReMap(ppath,  picode->args[picode->dest]);
	ifi->NewDest_Phys = ifi->args[picode->dest];
      }
    }

    if (ppath->instwin->enqueue(ifi) == -1) {
      fatal("Path instwin overflow" << endl);
    }
    return ifi;
  }
}



static jmp_buf issue_restart_jmp_buf;

int handle_signals_for_issue(int sig)
{
  if (AINT_in_issue) {
    long ieee_fp_control;
    long ieee_state_at_signal[3];
    getsysinfo(GSI_IEEE_FP_CONTROL, (caddr_t)&ieee_fp_control, sizeof(ieee_fp_control), 0, 0, 0);
    getsysinfo(GSI_IEEE_STATE_AT_SIGNAL, 
	       (caddr_t)&ieee_state_at_signal, sizeof(ieee_state_at_signal), 0, 0, 0);

    cerr << "handle_signals_for_issue!\n"
	 << "    - ieee_fp_control = " << (void *) ieee_fp_control
	 << "    - ieee_state_at_signal = [" << ieee_state_at_signal[0]
	 << ", " << ieee_state_at_signal[1]
	 << ", " << ieee_state_at_signal[2] << "]" << endl;

    longjmp(issue_restart_jmp_buf, sig);
  } else {
    /* oops -- simulator itself caught a signal -- fatal! */
    fatal("handle_signals_for_issue: caught signal " <<  sig << endl);
  }
  return 0;
}




void AINT_issue(inflight_inst_ptr ifi)
{
  path_ptr ppath = ifi->ppath;
  thread_ptr pthread = ppath->pthread;

  int sig;

  AINT_in_issue = 1;
  if ((sig = setjmp(issue_restart_jmp_buf)) == 0) {
    icode_ptr picode = ifi->picode;
    icode_ptr new_picode;
    if (picode) {
      /* This is the time to copy the old destination register value
	 into the new one */
      if (picode->dest < MaxArgs) {
        pthread->Reg[ifi->NewDest_Phys].Int64 = pthread->Reg[ifi->OldDest_Phys].Int64;
      }
      ifi->runstate = R_RUN;
      ifi->signal = 0;

      if (ifi->picode->memop_func) {

	// the event_write memop function updates the aint store buffer.

	ifi->picode->memop_func(ifi, ifi->picode, ifi->ppath, ifi->ppath->pthread);
      } else {

	new_picode = ifi->picode->action(ifi, ifi->picode, ifi->ppath, ifi->ppath->pthread);

	if (ifi->runstate == R_SIGNAL) {
	  AINT_in_issue = 0;
	  return;
	} else {
	  ifi->nextPC = new_picode ? new_picode->addr : (aint_addr_t) 0;
	  // picode = new_picode;
	}
      }
    }
  } else {
    // Arrived via longjmp

    // Look at this at commit time:
    ifi->runstate = R_SIGNAL;
    ifi->signal = sig;
  }
  AINT_in_issue = 0;
}


void AINT_do_read(inflight_inst_ptr ifi)
{
  path_ptr ppath = ifi->ppath;
  thread_ptr pthread = ppath->pthread;
  icode_ptr new_picode;

  ulong bitmask = ifi->bitmask;
  int size = ifi->picode->size;

  ifi->data = 0l;
  /* Read into register from storebuf and/or from cache */
  pthread->StoreBuffer->Lookup(ifi, bitmask);
  if (bitmask) {
    /* Read the rest of the word from the cache */
    ifi->data &= ~bitmask;
    if (ifi->paddr) ifi->data |= (* (long *) (ifi->paddr & QWORD_MASK) & bitmask);
  }
  

  /* At this point, we have the quad-word over the given address. We need to
   * shift it to ensure the required data starts at byte 0
   */
  if (size < 8) {
    int shift = ((ifi->vaddr & 0x7) * 8);
    ulong mask = (1ul << (size * 8)) - 1;
    int signpos = 1 << (size * 8 - 1);
    ifi->data = (ifi->data >> shift);
    /* now sign extend */
    if (0 &&ifi->data & signpos)
      ifi->data |= ~mask;
  }

  new_picode = ifi->picode->func(ifi, ifi->picode, ppath, pthread);
  ifi->nextPC = (new_picode ? new_picode->addr : (aint_addr_t) 0);

  /* Set ifi->done_memop to true */
  ifi->done_memop = 1;

  /* If inst committed, set ifi->picode to null */
  if (ifi->committed) {
    ifi->picode = NULL;
  }

}



void AINT_do_write(inflight_inst_ptr ifi)
{

  path_ptr ppath = ifi->ppath;
  thread_ptr pthread = ppath->pthread;
  icode_ptr new_picode;

  /* We have the quad-word over the given address. We need to
   * shift it to ensure the required data starts at byte 0
   */

  /* Write through to cache */
  new_picode = ifi->picode->func(ifi, ifi->picode, ppath, pthread);
  ifi->nextPC = (new_picode ? new_picode->addr : (aint_addr_t) 0);

  /* If we wrote to an address locked by someone, unlock it */
  for (int i=0; i<ThreadCount; i++) {
    if (Threads[i]->locked_addr == (ifi->vaddr >> LOCK_RANGE_SZ)) {
      Threads[i]->lock_flag = 0;
    }
  }
  /* Set ifi->done_memop to true */
  ifi->done_memop = 1;


  /* If inst committed, set ifi->picode to null; cleanup storeq */
  if (ifi->committed) {
    pthread->StoreBuffer->Remove(ifi);
    ifi->picode = NULL;


    informative("AINT_do_write: uncommitted store pc="
		<< (void *) ifi->picode->addr << ", fetchnum="
		<< (void *) ifi->fetchnum << ", vaddr="
		<< (void *) ifi->vaddr);

    /* potentially a load-store trap was missed, check it out: */
    /* fatal if any older (lower-fetchnum) loads to same address */
    /* Go down cbif until the end or first load that has done read */

    inflight_inst_ptr oifi;

    for (oifi = ppath->instwin->pred(ifi); oifi; oifi=ppath->instwin->pred(oifi)) {
      if (   (oifi->iflags & E_READ)
	  && ADDR_OVERLAP(oifi->vaddr, oifi->picode->size,  ifi->vaddr, ifi->picode->size)
	  && (!oifi->done_memop) ) {
	fatal("AINT_do_write: pc=" << (void *) ifi->picode->addr
	      << ", fetchnum=" << (void *) ifi->fetchnum
	      << ", vaddr=" << (void *) ifi->vaddr
	      << "\n\tThere is an older outstanding load to the same address" << endl);

	break;
      }
    }
  }



}




void AINT_commit(inflight_inst_ptr ifi)
{
  path_ptr ppath = ifi->ppath;
  thread_ptr pthread=ppath->pthread;
#if 0
  if (Debugger_tty_name != NULL)
    gdbserver_cycle();
#endif

  inflight_inst_ptr committed_ifi = ppath->instwin->dequeue();
  if (committed_ifi != ifi) {
    fatal("Path " << (void *) ppath->pathid
	  << ": Committing "<< (void *) ifi->picode->addr << "(" << ifi->fetchnum
	  << ") out-of-order; oldest inst has fetchnum " << committed_ifi->fetchnum << endl);
  }

  if (ifi->runstate == R_SIGNAL) {
    // should be fatal
    fatal("AINT_commit: " << pthread->pprocess->pid << "." << pthread->tid
	  << "died with signal " << ifi->signal
	  << "in/before " << (void *) ifi->picode->addr << ":"
	  << (void *) ifi->picode->instr << ":" << desc_table[ifi->picode->opnum].opname << endl);
  }

  // Release the old dest reg if we renamed it
  if (ifi->OldDest_Phys != ZERO_REGISTER) {
    assert (pthread->Is_InUse(ifi->OldDest_Phys));
    pthread->Clr_InUse(ifi->OldDest_Phys);
  } else {
    icode_ptr picode = ifi->picode;
    assert ( (picode->dest == MaxArgs) || (picode->args[picode->dest] == ZERO_REGISTER) );
  }


  if (ifi->iflags & E_MEM_REF) {
    if (!ifi->done_memop) {
      cerr << "Committing ifi without completing memop at " 
	   << (void *) ifi->picode->addr << ":" << (void *) ifi->picode->instr << endl;
    }
  }

  if (ifi->stb_update) pthread->StoreBuffer->Remove(ifi);

  // If this is the last instruction in the path, does the
  // path have a child? If so, promote that subtree.
  if (ppath->instwin->is_empty()) {
    assert( !(ppath->left && ppath->right) );
    path_ptr child_path = ppath->left ? ppath->left : ppath->right;
    if (child_path) {
      // Promote child path to root
      child_path->promote(ROOT_PATH_ID);

      // This path should now die since all insts have committed
      // FIXME: verify that there is no race condition here
      delete ppath;

    }
  }
  delete ifi;
}


// This kills the path subtree starting from the instruction
void AINT_kill(inflight_inst_ptr ifi)
{
  // For now, assume single-path

  // Start at the last inst in this path, and roll back
  path_ptr ppath = ifi->ppath;
  thread_ptr pthread = ppath->pthread;
  inflight_inst_ptr dead_ifi;

  // If there are children, kill them
  if (ppath->left) AINT_kill_path(ppath->left);
  if (ppath->right) AINT_kill_path(ppath->right);

  while (1) {
    dead_ifi = ppath->instwin->pop();
    icode_ptr picode = dead_ifi->picode;


    if (picode->dest < MaxArgs) {
      Physical_RegNum pr = dead_ifi->args[picode->dest];
      if (pr != ZERO_REGISTER) {
	pthread->Clr_InUse(pr);
	ppath->RegMap[picode->args[picode->dest]] = dead_ifi->OldDest_Phys;
	assert(ppath->pthread->Is_InUse(dead_ifi->OldDest_Phys));
	// REG(picode->dest) = dead_ifi->OldDest_Phys;
      }
    }

    // If this was a store, dequeue and delete store-queue update
    if (dead_ifi->stb_update) pthread->StoreBuffer->Remove(dead_ifi);

    if (dead_ifi == ifi) break;
    delete dead_ifi;

  }
  delete dead_ifi;

}


// This tells aint to use <alt_pc> to create a new path at this instruction
path_ptr
AINT_create_path(inflight_inst_ptr ifi, aint_addr_t start_pc)
{
  path_ptr ppath = ifi->ppath;
  if (ppath->left) {

    if (ppath->right) return (path_ptr) NULL;

    ppath->right = new path(RIGHTCHILD(ppath->pathid), ppath, ppath->pthread, start_pc);
    return ppath->right;
  } else {
    ppath->left = new path(LEFTCHILD(ppath->pathid), ppath, ppath->pthread, start_pc);
    return ppath->left;
  }
}

// This tells aint to start a new root path because the old was killed by the backend
path_ptr AINT_create_path(int tid, aint_addr_t start_pc)
{
  thread_ptr pthread = Threads[tid];

  if (pthread->path_tree == (path_ptr) NULL) {
    pthread->path_tree = new path(ROOT_PATH_ID, (path_ptr) NULL, pthread, start_pc);
  }

  assert(pthread->path_tree->instwin->is_empty());
  return pthread->path_tree;
}


// Create a path as the child of specified path - all checks OFF (for now)
path_ptr AINT_create_path(path_ptr ppath, aint_addr_t start_pc)
{
  assert(ppath != NULL);
  // assert(!(ppath->pathid >> 63));

  if (ppath->pathid >> 63) {
    // pathid overflow - depth exceeds 64
    return (path_ptr) NULL;
  }

  if (ppath->left) {

    if (ppath->right) return (path_ptr) NULL;

    ppath->right = new path(RIGHTCHILD(ppath->pathid), ppath, ppath->pthread, start_pc);
    return ppath->right;
  } else {
    ppath->left = new path(LEFTCHILD(ppath->pathid), ppath, ppath->pthread, start_pc);
    return ppath->left;
  }
}


// Kills the path subtree rooted at pathid
// Recursive post-order traversal
void AINT_kill_path(path_ptr ppath)
{
  if (ppath->left) AINT_kill_path(ppath->left);
  if (ppath->right) AINT_kill_path(ppath->right);

  while (!ppath->instwin->is_empty()) {
    inflight_inst_ptr ifi = ppath->instwin->pop();
    Arg dest = ifi->picode->dest;
    if (dest < MaxArgs) {
      if (ifi->args[dest] != ZERO_REGISTER) {
	ppath->pthread->Clr_InUse(ifi->args[dest]);
      }
    }
    ppath->pthread->StoreBuffer->Remove(ifi);
    delete ifi;  // FIXME: This should remove the storebuffer update, if any
  }

  // Let the parent know we're dead
  if (ppath->parent) {
    if (ppath->parent->left == ppath) ppath->parent->left = (path_ptr) NULL;
    if (ppath->parent->right == ppath) ppath->parent->right = (path_ptr) NULL;
  }

  // delete the path, now that the instwin has been freed
  delete ppath;
}


// The event_* functions - executed for special mem-op events

OP(event_read)
{
  aint_addr_t vaddr;
  unsigned long paddr;

  /* The vaddr calculated is the address passed on by the instruction decoder.
   * The actual address referenced may be different, as in case of unaligned
   * loads (and stores)
   */
  vaddr = REG(RB) + picode->immed;
  if (picode->opnum == ldq_u_opn) {
    vaddr = BASE2ROUNDDOWN(vaddr, 8);
  }

  ifi->vaddr = vaddr;

  paddr = (ulong) pthread->pprocess->PMAP(vaddr);
  ifi->pte = pthread->pprocess->tlb->Translate(ifi->vaddr, (long *) NULL);

  ifi->paddr = paddr;


  /* Determine fe_vector and fe_mask */
  if (ifi->iflags & E_SYNC) {
    if (paddr) {
      ifi->fe_vector = ifi->pte->fe_vector + (TB_OFFSET(ifi->vaddr)>>9);
      ifi->fe_mask = ( 1ul << ((ifi->vaddr>>3)&(0x3f)) ); /* 0x3f isolates 6 bits (0..63) */
    }
  }

  /* Calculate the byte-mask */
  switch (picode->size) {
  case 1:
  case 2:
  case 4:
    ifi->bitmask = (1L << (8 * picode->size)) - 1;
    if (0) informative((void *) picode->addr << ": vaddr=" << (void *) ifi->vaddr 
		       << ", size=" << picode->size << ", bitmask=" << (void *) ifi->bitmask << endl);
    ifi->bitmask = ifi->bitmask << ((ifi->vaddr & 0x7)<<3);
    if (0) informative("\t bitmask2=" << ifi->bitmask << endl);
    break;
  case 8:
    ifi->bitmask = ~(0ul);
    break;
  default:
    fatal(pthread->pprocess->pid << "." << pthread->tid << ": " << (void *) picode->addr
	  << ": invalid data size (" << picode->size << " for memop" << endl);
  }

  if (paddr) {
    if (picode->size == 8) {
      if (paddr & 7)
	fprintf(stderr, "event_read: unaligned access pid %d thread %d paddr 0x%lx vaddr 0x%lx pc 0x%lx inst %08x\n", 
	       pthread->pprocess->pid, pthread->tid, paddr, vaddr, picode->addr, picode->instr);
      /* pevent->data = * (long *) paddr; */
    }
  } else {
    ifi->runstate = R_SIGNAL;
    ifi->signal = SIGSEGV;
  }

#if 0
  informative("event_read: vaddr " << (void *) vaddr << " paddr " << (void *)paddr << endl);
#endif

  return picode->next;
  /* return value doesn't matter */


}



OP(event_write)
{
  aint_addr_t vaddr;
  long paddr;
  tb_entry *pte;

  vaddr = REG(RB) + picode->immed;
  if (picode->opnum == stq_u_opn) vaddr = BASE2ROUNDDOWN(vaddr, 8);

  ifi->vaddr = vaddr;

  paddr = (ulong) pthread->pprocess->PMAP(vaddr);
  ifi->pte = pthread->pprocess->tlb->Translate(ifi->vaddr, (long *) NULL);

  /* Determine fe_vector and fe_mask */
  if (paddr) {
    ifi->fe_vector = ifi->pte->fe_vector + (TB_OFFSET(ifi->vaddr)>>9);
    ifi->fe_mask = ( 1ul << ((ifi->vaddr>>3)&(0x3f)) ); /* 0x3f isolates 6 bits (0..63) */
  }

  /* Convert register value to memory word */
  if (picode->iflags & E_FLOAT) {
    aint_cvt_fp_mem (picode->opnum, FP(RA), &ifi->data);
  } else {
    ifi->data = REG(RA);
  }

  /* Calculate the byte-mask */
  switch (picode->size) {
  case 1:
  case 2:
  case 4:
    ifi->bitmask = (1L << (8 * picode->size)) - 1L;
    ifi->bitmask = ifi->bitmask << ((ifi->vaddr & 0x7)<<3);
    break;
  case 8:
    ifi->bitmask = ~(0ul);
    break;
  default:
    fatal(pthread->pprocess->pid << "." << pthread->tid << ": " << (void *) picode->addr
	  << ": invalid data size (" << picode->size << ") for memop" << endl);

  }
  ifi->paddr = paddr;

#if TRACE_EXEC
  informative("event_write: data is " << ifi->data);
  informative(" vaddr " << (void *) vaddr << " paddr " << (void *)paddr << endl);
#endif

  /* Update the store buffer using ifi->data, size, bitmask */
  pthread->StoreBuffer->Update(ifi);

  return picode->next;
  /* return value not used */


}


OP(event_load_locked)
{
  return event_read(ifi, picode, ppath, pthread);
}

/* There's a problem with this implementation of event_store_conditional.
 * event_write, which is called by event_store_conditional, updates the store-buffer,
 * which may be used by a newer read to the same address. If the store-conditional
 * fails, then the load has the incorect value, and should either be trapped, or
 * should stall on the store-buffer read. At this time, we leave it up to the backend to
 * ensure correctness.
 */
OP(event_store_conditional)
{
  return event_write(ifi, picode, ppath, pthread);
}



OP(event_memory_barrier)
{
  // FIXME: how should this event be handled?
  // Should the simulator know that this is a barrier?
  // Remove this from the memop list?
  // return picode->func(ifi, picode, ppath, pthread);
  return (icode_ptr) NULL;
}

OP(event_sim_user)
{
#if 0
  event_ptr pevent;

  pevent = pthread->pevent;
  pevent->iaddr = picode->addr;
  pevent->arg1 = MapReg(A0);
  pevent->arg2 = MapReg(A1);
  pevent->arg3 =  MapReg(A2);
  pevent->arg4 = MapReg(A3);
  pthread->ufunc = sim_user;
  SLEEP(pthread);
#endif

  return(picode->next);
}

