/* 
 * Routines for loading executables and shared objects
 */

#include <iostream.h>
#include <fstream.h>
#include <strstream.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/mman.h>

#include <sys/time.h>
#include <sys/resource.h>

#include <time.h>
#include <sys/utsname.h>
#include <string.h>
#include <assert.h>

#if defined(__LANGUAGE_C__) || defined(__cplusplus)
#else
#error bozo compilers apply here
#endif

#include <elf_abi.h>
#include <elf_mips.h>

#include "aint.h"
#include "icode.h"
#include "opcodes.h"
#include "globals.h"

#define USE_ATOM_OUTPUT 0

/* prototypes for functions */

#include "protos.h"

#include "bfd.h"
extern "C" {long bfd_elf_hash(const unsigned char *);}

static int loader_initialized = 0;

static void decode_opnum(icode_ptr picode, unsigned instr);

#if USE_ATOM_OUTPUT
static void read_bb_info(process_ptr process, const char *Objname);
#endif


static const char *loader_target = "ecoff-littlealpha";

static void loader_init ()
{
  bfd_init();

}



int emulate_sbin_loader = 0; 

#define MAX_N_OBJECT_FILES 256
typedef struct elf_hash_s {
  Elf32_Word nbuckets;
  Elf32_Word nchains;
  Elf32_Word *bucket;
  Elf32_Word *chain;
} elf_hash_t;

typedef struct gotinfo_s {
  Elf32_Word local_gotno;
  Elf32_Word gotsym;
} gotinfo_t;

typedef struct object_file_s {
  const char *filename;
  char loaded[MAX_NPROCS];

  int flags; /* copy of abfd->flags */
  int n_libs_needed; 
  Elf32_Lib *liblist;

  char *dynstr;
  Elf32_Sym *dynsym;
  elf_hash_t hash;

  gotinfo_t *gotinfo;
  int ngots;
  bfd *abfd;
  Elf32_Addr stack_end;

  /* other info from .dynamic section */
  Elf32_Addr dt_pltgot;
  Elf32_Addr dt_init;
  Elf32_Addr dt_fini;
  Elf32_Word dt_mips_time_stamp;
  Elf32_Word dt_mips_ichecksum;
} object_file_t;

object_file_t object_file[MAX_N_OBJECT_FILES];
int n_object_files = 0;

object_file_t *
object_file_add(const char *name)
{
  int i;
  for (i = 0; i < n_object_files; i++) {
    if (strcmp(object_file[i].filename, name) == 0)
      return &object_file[i];
  }
  object_file[n_object_files].filename = name;
  return &object_file[n_object_files++];
} 

object_file_t *
object_file_get(const char *name)
{
  int i;
  for (i = 0; i < n_object_files; i++) {
    if (strcmp(object_file[i].filename, name) == 0)
      return &object_file[i];
  }
  return NULL;
} 

int 
object_file_open(object_file_t *pof)
{
  bfd *abfd;
  if (pof->abfd)
    return -1;

  informative("object_file_open: " << pof->filename << endl);

  if ((abfd = pof->abfd = bfd_openr(pof->filename, loader_target)) == NULL) {
    // char msg[1024];
    ostrstream msg;
    msg << "object_file_open(" << pof->filename << ")";
    // sprintf(msg, "object_file_open(%s)", pof->filename);
    bfd_perror(msg.str());
    delete (char *) msg.str();
    return -1;
  } 
  if (!bfd_check_format(pof->abfd, bfd_object))
    bfd_perror("load_object_file - check_format");

  /* save the flags */
  pof->flags = abfd->flags;

  if (emulate_sbin_loader) {
    if (!pof->dynstr) {
      asection *dynstr_section = bfd_get_section_by_name(abfd, ".dynstr");
      if (dynstr_section != NULL) {
	int size = bfd_section_size(abfd, dynstr_section);
	pof->dynstr = new char[size];
	bfd_get_section_contents(abfd, dynstr_section, (char*)pof->dynstr, 0, size);
      }
    }
    if (!pof->dynsym){
      asection *dynsym_section = bfd_get_section_by_name(abfd, ".dynsym");
      if (dynsym_section != NULL) {
	int size = bfd_section_size(abfd, dynsym_section);
	pof->dynsym = (Elf32_Sym *)malloc(size);
	bfd_get_section_contents(abfd, dynsym_section, (char*)pof->dynsym, 0, size);
      }
    }

    if (!pof->hash.nbuckets) {
      asection *hash_section = bfd_get_section_by_name(abfd, ".hash");
      if (hash_section != NULL) {
	int size = bfd_section_size(abfd, hash_section);
	Elf32_Word *rawhash = (Elf32_Word *)malloc(size);
	bfd_get_section_contents(abfd, hash_section, (char*)rawhash, 0, size);
	pof->hash.nbuckets = rawhash[0];
	pof->hash.nchains = rawhash[1];
	pof->hash.bucket = &rawhash[2];
	pof->hash.chain = &rawhash[2+pof->hash.nbuckets];
	informative("hash_section size = " << size << endl);
      }
    }
  }

  return 0;
}

void
object_file_close(object_file_t *pof)
{
  bfd_close(pof->abfd);
  pof->abfd = NULL;
}

const char *object_file_dynstr(object_file_t *pof, Elf32_Word dynstrIndex)
{
  if (pof->dynstr)
    return pof->dynstr + dynstrIndex;
  else 
    return NULL;
}

Elf32_Sym *
object_file_dynsym_from_index(object_file_t *pof, Elf32_Word dynsymIndex)
{
  if (pof->dynsym)
    return pof->dynsym + dynsymIndex;
  else 
    return NULL;
}


Elf32_Sym *
object_file_dynsym_from_name(object_file_t *pof, const char *name)
{
  Elf32_Word nbuckets = pof->hash.nbuckets;
  Elf32_Word *bucket = pof->hash.bucket;
  Elf32_Word *chain = pof->hash.chain;
  long h = bfd_elf_hash((const unsigned char *) name);
  int b = h % nbuckets;
  int i = bucket[b]; 
  Elf32_Sym *sym = object_file_dynsym_from_index(pof, i);
  const char *symname = object_file_dynstr(pof, sym->st_name);
  if (strcmp(symname, name) == 0) {
    return sym;
  } else {
    while ((i = chain[i]) != STN_UNDEF) {
      sym = object_file_dynsym_from_index(pof, i); 
      symname = object_file_dynstr(pof, sym->st_name);
      if (strcmp(symname, name) == 0) 
	return sym;
    }
    return NULL;
  }
}



void
ofse_fcn (bfd *abfd, asection *section, object_file_t *pof)
{
  if (section->vma == 0)
    return;
  if (pof->stack_end == 0)
    pof->stack_end = section->vma;
  if (section->vma < pof->stack_end)
    pof->stack_end = section->vma;
}

void
object_file_info(const char *objfile_name, 
		 aint_addr_t *pstack_end,
		 int *pdynamic)
{
  object_file_t *pof = object_file_get(objfile_name);
  if (!pof) {
    pof = object_file_add(objfile_name);
    object_file_open(pof);
    informative("object_file_info: " << objfile_name << " flags= " << pof->abfd->flags << endl);
    bfd_map_over_sections(pof->abfd, 
			  (void (*)(bfd*,asection*,void*))ofse_fcn,
			  (void*)pof);
    object_file_close(pof);
  }
  *pstack_end = pof->stack_end;
  *pdynamic = (pof->flags & DYNAMIC);
}



void object_file_load(process_ptr pprocess, object_file_t *pof);
/* load_section is funky because it is called from bfd_map_over_sections() */

struct ofl_state {
  object_file_t *pof;
  process_ptr pprocess;
};

static void load_section(bfd *abfd, asection *section, struct ofl_state *);
static void load_code_section(process_ptr pprocess, object_file_t *pof, asection *section);
static void load_liblist_section(process_ptr pprocess, object_file_t *pof, asection *section);

static void load_dynamic_section(process_ptr pprocess, object_file_t *pof, asection *section);
static void adjust_got_section(process_ptr pprocess, object_file_t *pof);

static void analyze_code_segment(process_ptr pprocess, ulong start, ulong size);


/* load_object_file (process, filename)
   uses BFD to read the executable to which filename points
   and readies process for execution.

   load_object_file also parses all the instructions contained in filename.
   */
void
load_object_file(process_ptr pprocess, const char *exename)
{
  object_file_t *pof;
  bfd *abfd;
  if (!loader_initialized)
    loader_init();

  pof = object_file_add(exename);
  object_file_open(pof);
  abfd = pof->abfd;
  assert(abfd->flags & EXEC_P);
  
  pprocess->entry_point = abfd->start_address;
  pprocess->initial_gp = bfd_ecoff_get_gp_value(abfd);
  pprocess->brk = 0;

  object_file_load(pprocess, pof);

  if (emulate_sbin_loader)
    adjust_got_section(pprocess, pof);

  object_file_close(pof);

  informative("load_object_file: entry point is " << (void *) pprocess->entry_point << endl);
  informative("load_object_file: gp is " << (void *) pprocess->initial_gp << endl);
  informative("load_object_file: found " << pprocess->n_instructions << " instructions" << endl);

}

void load_library(process_ptr pprocess, const char *libname)
{
  object_file_t *pof;
  bfd *abfd;
  if (!loader_initialized)
    loader_init();

  pof = object_file_add(libname);
  object_file_open(pof);
  abfd = pof->abfd;
  assert(abfd->flags & EXEC_P);
  
  object_file_load(pprocess, pof);

  informative("load_library: entry point is " << (void *) abfd->start_address << endl);
  informative("load_library: gp is " << (void *) bfd_ecoff_get_gp_value(abfd) << endl);

  object_file_close(pof);

}

void object_file_load(process_ptr pprocess, object_file_t *pof)
{
  bfd *abfd = pof->abfd;
  struct ofl_state ofl = { pof, pprocess };

  informative("load_object_file: (" << pof->filename << ")" << endl);
  informative("load_object_file: flags are " << (void *) abfd->flags << endl);

  bfd_map_over_sections(abfd, 
			(void (*)(bfd *, asection *, void *))load_section,
			(void *)&ofl);
  

  pof->loaded[pprocess->pid] = 1;

  if (emulate_sbin_loader) {
    int i;
    for (i = 0; i < pof->n_libs_needed; i++) {
      // char filename[1024];
      ostrstream filename;
      const char *name = object_file_dynstr(pof, pof->liblist[i].l_name);
      object_file_t *lpof;

      filename << "/usr/shlib/" << name;
      // sprintf(filename, "/usr/shlib/%s", name);
      lpof = object_file_add( strdup(filename.str()) );
      delete (char *) filename.str();

      if (!lpof->loaded[pprocess->pid]) {
	load_library(pprocess, lpof->filename);
      }
    }
  }

#if USE_ATOM_OUTPUT
  read_bb_info(process, exename);
#endif

}


static void load_code_section(process_ptr pprocess, object_file_t *pof, asection *section)
{
  ulong start = section->vma;
  ulong size = section->_raw_size;
  pprocess->decode_text_region(start, size);

}




static void load_section(bfd *abfd, asection *section, struct ofl_state *state)
{
  object_file_t *pof = state->pof;
  process_ptr pprocess = state->pprocess;
  flagword flags = section->flags;
  ulong start = section->vma;
  ulong size = section->_raw_size;

  if (flags & SEC_NEVER_LOAD) {
    /* do nothing */
    informative("load_section(" << section->name << ": start ="
		<< (void *)start <<", size=" << (void *)size <<", flags="
		<< (void *) flags << " -- do nothing" << endl);

    //    informative("load_section(%s: start=%lx, size=%lx, flags=%x -- do nothing \n",
    // section->name, start, size, flags);
  } else if (flags & SEC_ALLOC) {
    int prot = PROT_READ;

    informative("load_section(" << section->name << ": start=" << (void *) start
		<< ", size=" << (void *) size << ", flags=" << (void *)flags);

    if (!(flags & SEC_READONLY) && (flags & SEC_DATA))
      prot |= PROT_WRITE;
    if (flags & SEC_CODE) {
      prot |= PROT_EXEC;
    }

    pprocess->add_segment(start, size, prot, (flags & SEC_CODE) ? MAP_SHARED : MAP_PRIVATE);
    /* if (!strcmp(section->name, ".bss"))  */
    if (strcmp(section->name, ".bss")) 
      pprocess->brk = start+size;
    if (strcmp(section->name, ".liblist") == 0)
      load_liblist_section(pprocess, pof, section);
    if (strcmp(section->name, ".dynamic") == 0)
      load_dynamic_section(pprocess, pof, section);

    if (flags & (SEC_LOAD|SEC_HAS_CONTENTS)) {
      char *buf = new char[size];
      if (!bfd_get_section_contents(abfd, section, buf, 0, size)) {
	bfd_perror("load_section - get_section_contents");
      }
      /* now copy the contents */
      pprocess->memcpy_to_object_space(start, buf, size);
      free(buf);
      if (flags & SEC_CODE) {
	load_code_section(pprocess, pof, section);
	if (0)
	analyze_code_segment(pprocess, section->vma, section->_raw_size);
      }
    } else {
      /* zero-filled section: do nothing */
    }
  }
}



int 
segment::contains(aint_addr_t addr)
{
  return ((start <= addr) && (addr <= end));
}

segment_t *
process::add_segment(aint_addr_t start, ulong size, int prot, int flags)
{
  int i;
  int is_new = 0;
  segment_t *segment = NULL;
  ulong rounded_start = BASE2ROUNDDOWN(start, (ulong)TB_PAGESIZE);
  ulong rounded_end = BASE2ROUNDUP(start+size, (ulong)TB_PAGESIZE)-(ulong)1;
  ulong rounded_size = rounded_end - rounded_start+(ulong)1;

  informative("process_add_segment [" << (void *) start <<", " << (void *) (start+size)
	      << "] [" << (void *) flags << "]" << endl);
  informative(" \t => [" << (void *) rounded_start <<", "
	      << (void *) rounded_end << ", " << (void *) rounded_size << "]" << endl);

  /* check for existing segment */
  for (i = 0; i < n_segments; i++) {
    if (segments[i].contains(rounded_start) || segments[i].contains(rounded_end)) {
      segment = &segments[i];
      break;
    }
  }

  if (segment == NULL) {
    is_new = 1;
    n_segments++;
    segments = (segment_t *) realloc(segments, n_segments * sizeof(segment_t));
    segment = &segments[n_segments-1];
    segment->start = rounded_start;
    segment->end = rounded_end;
    segment->size = rounded_size;
    segment->prot = prot;
    segment->flags = flags;
    informative("\t added new segment" << endl);
  } else {
    if (segment->start > rounded_start)
      segment->start = rounded_start;
    if (segment->end < rounded_end)
      segment->end = rounded_end;
    segment->size = 
      (ulong)((ulong)segment->end - (ulong)segment->start + (ulong)1);
  }
  informative(" \t => [" << (void *) segment->start << ", " << (void *) segment->end
	      << ", " << (void *) segment->size << "] " 
	      << (is_new ? "new" : "extended") << " segment "<< i << ")" << endl);
  segment->prot |= prot;
  return segment;
}

segment_t *
process::find_segment(aint_addr_t addr)
{
  int i;
  /* check for containing segment */
  for (i = 0; i < n_segments; i++) {
    if (segments[i].contains(addr)) {
      return &segments[i];
    }
  }
  /* no access */
  return NULL;
}



#define ALIGNED(v_, type_) ((((long)(v_)) & (sizeof(type_)-1)) == 0)


void 
process::memcpy_from_object_space(void *aint_ptr, const aint_addr_t object_ptr, size_t size)
{
  ulong i;
  /* optimize this more later */
  if (ALIGNED(size, long) && ALIGNED(object_ptr, long) && ALIGNED(aint_ptr, long)) {
    for (i=0; i<size; i += sizeof(long)) {
      *(long*)(aint_ptr + i) = * (long *) addr2phys(object_ptr + i, NULL);
    }
  } else if (ALIGNED(size, int) && ALIGNED(object_ptr, int) && ALIGNED(aint_ptr, int)) {
    for (i=0; i<size; i += sizeof(int)) {
      *(int*)(aint_ptr + i) = * (int *) addr2phys(object_ptr + i, NULL);
    }
  } else {
    for (i=0; i<size; i += sizeof(char)) {
      *(char*)(aint_ptr + i) = * (char *) addr2phys(object_ptr + i, NULL);
    }
  }
}

void
process::memcpy_to_object_space(aint_addr_t object_ptr, void *aint_ptr, size_t size)
{
  ulong i;
  /* optimize this more later */
  if (ALIGNED(size, long) && ALIGNED(object_ptr, long) && ALIGNED(aint_ptr, long)) {
    for (i=0; i<size; i += sizeof(long)) {
      * (long *) addr2phys(object_ptr + i, NULL) = * (long*) (aint_ptr + i);
    }
  } else if (ALIGNED(size, int) && ALIGNED(object_ptr, int) && ALIGNED(aint_ptr, int)) {
    for (i=0; i<size; i += sizeof(int)) {
      * (int *) addr2phys(object_ptr + i, NULL) = * (int*) (aint_ptr + i);
    }
  } else {
    for (i=0; i<size; i += sizeof(char)) {
      * (char *) addr2phys(object_ptr + i, NULL) = * (char*) (aint_ptr + i);
    }
  }
}


size_t
process::strlen_from_object_space(aint_addr_t object_ptr)
{
  long i;
  size_t len = 0;
  /* optimize this later */
  for (i=0; ; i++) {
    char c = * (char *) addr2phys(object_ptr + i, NULL);
    if (c == '\0') 
      break;
    else
      len++;
  }
  return len;
}

char *
process::strdup_from_object_space(aint_addr_t object_ptr)
{
  long i;
  size_t size = 1024;
  char *aint_ptr = (char *) malloc(size);
  size_t len = 0;

  /* optimize this later */
  for (i=0; ; i++) {
    char c = * (char *) addr2phys(object_ptr + i, NULL);
    aint_ptr[i] = c;
    if (c == '\0') {
      break;
    } else {
      len ++;
      if (len == size) {
	size *= 2;
	aint_ptr = (char *) realloc(aint_ptr, size);
      }
    }
  }
  return aint_ptr;
}


void
process::strcpy_from_object_space(char *aint_ptr, aint_addr_t object_ptr)
{
  long i;
  /* optimize this later */
  for (i=0; ; i++) {
    char c = * (char *) addr2phys(object_ptr + i, NULL);
    aint_ptr[i] = c;
    if (c == '\0') break;
  }
}

void
process::strcpy_to_object_space(aint_addr_t object_ptr, const char *aint_ptr)
{
  long i;
  /* optimize this later */
  for (i=0; ; i++) {
    char c = aint_ptr[i];
    * (char *) addr2phys(object_ptr + i, NULL) = c;
    if (c == '\0') break;
  }
}




/* Decode the given instruction: determine the aint opnum for the inst. */


static
#ifdef __GNUC__
  inline
#else
#  pragma inline(decode_opnum_group_entry)
#endif
int decode_opnum_group_entry(group_t *group, int function)
{
  int i;
  for (i=0; group[i].opnum != 0; i++) {
    if (group[i].opkey == function)
      return group[i].opnum;
  }
  return default_opnum;
}

static void
decode_opnum(icode_ptr picode, unsigned instr)
{
  int opcode;
  union alpha_instruction uinst;

  picode->instr = instr;

  uinst = * (union alpha_instruction *) &instr;

  opcode = uinst.common.opcode;

  /* Now, use the opcode and other fields to zero in on the actual
   * instruction.
   */

  switch (opcode) {
  case op_call_pal:
    picode->opnum = decode_opnum_group_entry(pal_group, uinst.pal_format.function);
    break;
  case op_inta:
    picode->opnum = decode_opnum_group_entry(inta_group, uinst.o_format.function);
    break;
  case op_intl:
    picode->opnum = decode_opnum_group_entry(intl_group, uinst.o_format.function);
    break;
  case op_intm:
    picode->opnum = decode_opnum_group_entry(intm_group, uinst.o_format.function);
    break;
  case op_ints:
    picode->opnum = decode_opnum_group_entry(ints_group, uinst.o_format.function);
    break;
  case op_fltv:
    picode->opnum = decode_opnum_group_entry(fltv_group, uinst.f_format.function);
    break;
  case op_flti:
    picode->opnum = decode_opnum_group_entry(flti_group, uinst.f_format.function);
    break;
  case op_fltl:
    picode->opnum = decode_opnum_group_entry(fltl_group, uinst.f_format.function);
    break;
  case op_opc14:
    picode->opnum = decode_opnum_group_entry(opc14_group, uinst.o_format.function);
    break;
  case op_opc1c:
    picode->opnum = decode_opnum_group_entry(opc1c_group, uinst.o_format.function);
    break;

  case op_misc:
    picode->opnum = decode_opnum_group_entry(misc_group, uinst.m_format.memory_displacement);
    break;

  case op_jsr:
    picode->opnum = decode_opnum_group_entry(jsr_group, uinst.j_format.function);
    break;

  default:
    picode->opnum = decode_opnum_group_entry(defgroup, uinst.common.opcode);
  }
}

/* Decode an instruction: store information in an icode struct */

void
decode_instr(process_ptr pprocess, aint_addr_t textaddr, unsigned instr, 
	     icode_ptr picode, icode_ptr next_picode)
{

  int opcode;
  union alpha_instruction uinst;
  op_fptr simfunc;

  picode->next = next_picode;
  picode->addr = textaddr;
  picode->cycles = 1;

  picode->instr = instr;
  uinst = * (union alpha_instruction *) &instr;

  picode->dest = MaxArgs;
  picode->args[RA] = picode->args[RB] = picode->args[RC] = ZERO_REGISTER;

  /* Get the aint operation number */

  opcode = uinst.common.opcode;
  decode_opnum(picode, instr);

  /* At this time, we know the opnum, so we can access information stored
   * in the op_desc desc_table. Information desired is the function pointer,
   * the instruction type etc.
   */
  picode->func = desc_table[picode->opnum].func;
  picode->iflags = desc_table[picode->opnum].iflags;
  
#if !USE_ATOM_OUTPUT
  picode->iflags |= E_BB_FIRST|E_BB_LAST;
#endif

  /*
   * Decode instruction arguments
   */

  switch (OpFormat[opcode]) {
  case PAL:
    picode->immed = uinst.pal_format.function;
    break;

  case Reserved:
    break;

  case Memory:
    picode->immed = uinst.m_format.memory_displacement;
    /*
     * Sign-extend the immed field if opcode is not op_misc
     */
    if (opcode != op_misc) {
      picode->immed = (picode->immed | ((picode->immed>>15) ? ~0xffff : 0));
      if (picode->opnum == ldah_opn) picode->immed = picode->immed << 16;
    }
    picode->args[RB] = uinst.m_format.rb;
    /*
     * If this is a load, redirect loads into the zero reg.
     * Also, the m_format is being used to handle j_format,
     * so look for those...
     */
    switch(uinst.m_format.opcode) {
    case op_ldf: 
    case op_ldg: 
    case op_lds: 
    case op_ldt: 
      picode->dest = RA;
      picode->args[RA] = FP_REDIRECT_Z(uinst.m_format.ra);
      break;
    case op_stf:
    case op_stg:
    case op_sts:
    case op_stt:
      picode->args[RA] = FP_REDIRECT_Z(uinst.m_format.ra);
      break;
    case op_ldq_u: 
    case op_ldl: 
    case op_ldq: 
    case op_ldl_l: 
    case op_ldq_l: 
    case op_ldq_s:
    case op_stl_c:
    case op_stq_c:
    case op_stq_s:
    case op_jsr:
    case op_lda:
    case op_ldah:
      picode->dest = RA;
      /* picode->args[RA] = ZERO_REDIRECT(uinst.m_format.ra); */
      picode->args[RA] = uinst.m_format.ra;
      break;
    default:
      picode->args[RA] = uinst.m_format.ra;
      break;
    }
    break;

  case Operate:
#ifdef USE_FAST_FUNC
    picode -> op_form = uinst.o_format.form;
#endif
    if (!uinst.o_format.form) {
      picode->args[RA] = uinst.o_format.ra;
      picode->args[RB] = uinst.o_format.rb;
      /* picode->args[RC] = ZERO_REDIRECT(uinst.o_format.rc); */
      picode->args[RC] = uinst.o_format.rc;
      picode->literal = 0xffff;
      picode->immed = uinst.o_format.function;
      /* fprintf(stderr, "<%d> r%d, r%d, r%d\n", picode->immed, picode->args[RA],
	 picode->args[RB], picode->args[RC]); */
    }
    else {
      picode->args[RA] = uinst.l_format.ra;
      /* picode->args[RC] = ZERO_REDIRECT(uinst.o_format.rc); */
      picode->args[RC] = uinst.o_format.rc;
      picode->literal = uinst.l_format.literal;
      picode->immed = uinst.l_format.function;
      /* fprintf(stderr, "<%d> r%d, #0x%x, r%d\n", picode->immed,
	 picode->args[RA], picode->literal, picode->args[RC]); */
    }
    picode->dest = RC;
    break;
  case Branch: {
    ulong targetaddr;
    picode->immed = uinst.b_format.branch_displacement;
    picode->immed = (picode->immed | ((picode->immed >> 20) ? ~0x1fffff : 0));
    targetaddr = picode->addr + (picode->immed<<2) + 4;

    if (pprocess->find_segment(targetaddr))
      picode->target = pprocess->addr2iphys(targetaddr, NULL);

    switch (uinst.b_format.opcode) {
    case op_br:
    case op_bsr:
      /* picode->args[RA] = ZERO_REDIRECT(uinst.b_format.ra); */
      picode->args[RA] = uinst.b_format.ra;
      picode->dest = RA;
      break;
    default:
      picode->args[RA] = (picode->iflags & E_FLOAT) ? FP_REDIRECT(uinst.b_format.ra): uinst.b_format.ra;
      break;
    }
    /* fprintf(stderr, "r%d, 0x%x\n", picode->args[RA], picode->immed); */
  } break;
  case Float:
    picode->args[RA] = FP_REDIRECT_Z(uinst.f_format.fa);
    picode->args[RB] = FP_REDIRECT_Z(uinst.f_format.fb);
    picode->args[RC] = FP_REDIRECT_Z(uinst.f_format.fc);
    picode->immed = uinst.f_format.function;
    picode->dest = RC;
    picode->iflags |= E_FLOAT;
    /* fprintf(stderr, "<%d> f%d, f%d, f%d\n", picode->immed, picode->args[RA],
       picode->args[RB], picode->args[RC]); */
    break;
  default:
    break;
  }
  /* At this time, we know the opnum, so we can access information stored
   * in the op_desc desc_table. Information desired is the function pointer,
   * the instruction type etc.
   */
  picode->func = desc_table[picode->opnum].func;
  picode->iflags = desc_table[picode->opnum].iflags;

#if !USE_ATOM_OUTPUT
  picode->iflags |= E_BB_FIRST|E_BB_LAST;
#endif

    /*
     * if this is the address of the sim_user
     * procedure entry, replace with aint-space
     * sim_user() function
     */
    if ( picode -> addr == pprocess->sim_user_addr ) {
      picode->func = event_sim_user;
    }

    simfunc = (op_fptr) NULL;
    if (picode->iflags & E_READ)
      simfunc = event_read;
    else if (picode->iflags & E_WRITE)
      simfunc = event_write;
    /* FIXME: the ldx_l and stx_c are currently E_READ and E_WRITE resp.
        * that needs to be changed.
       
	else if (picode->iflags & E_LD_L)
	simfunc = event_load_locked;
	else if (picode->iflags & E_ST_C)
	simfunc = event_store_conditional;

      */
    // else if (picode->iflags & E_BARRIER)
    // simfunc = event_memory_barrier;

    /* for all the load and store ops, record size and ufunc in icode */

    if (picode->iflags & (E_MEM_REF|E_LOCK)) {

      switch (picode->opnum) {
      case ldb_opn:
      case stb_opn:
	picode->size = 1;
	break;
      case ldw_opn:
      case stw_opn:
	picode->size = 2;
	break;
      case ldl_opn:
      case stl_opn:
      case ldl_l_opn:
      case stl_c_opn:
      case ldf_opn:
      case lds_opn:
      case stf_opn:
      case sts_opn:
	picode->size = 4;
	break;
      case ldq_u_opn:
      case ldq_opn:
      case ldg_opn:
      case ldt_opn:
      case stq_u_opn:
      case stq_opn:
      case stg_opn:
      case stt_opn:
      case ldq_l_opn:
      case stq_c_opn:
      case ldq_s_opn:
      case stq_s_opn:
	picode->size = 8;
	break;
      default:
	break;
      }
    }

    if (simfunc) {
      picode->memop_func = simfunc;
      picode->iflags |= E_SPECIAL;
    }

}

void
process::decode_text_region(ulong start, ulong size)
{
  ulong i;
  icode_ptr picode = addr2iphys(start, NULL);

  informative("decode_text_region: startaddr=" << (void *) start 
	      << ", size=" << size << ", picode=" << (void *)picode<<endl);

  n_instructions += size/sizeof(int);

  for (i = 0; i < size; i += sizeof(int)) {
    ulong textaddr = start + i;
    ulong nextaddr = textaddr + sizeof(unsigned int);
    unsigned instr = *(unsigned int *)addr2phys(textaddr, NULL);
    icode_ptr next_picode = (nextaddr < (start + size)) ? addr2iphys(nextaddr, NULL) : (icode_ptr) NULL;

    if (picode)
      decode_instr(this, textaddr, instr, picode, next_picode);

    /* get ready for the next instruction */
    picode = next_picode;
  }
}



void
load_dynamic_section(process_ptr pprocess, object_file_t *pof, asection *section)
{
  bfd *abfd = pof->abfd;
  gotinfo_t tgotinfo[1024];
  Elf32_Word local_gotno_count = 0;
  Elf32_Word gotsym_count = 0;
  int size = bfd_section_size(abfd, section);
  Elf32_Dyn *dyncontents = (Elf32_Dyn*)malloc(size);
  Elf32_Dyn *dyn;
  Elf32_Word i;

  bfd_get_section_contents(abfd, section, (char*)dyncontents, 0, size);

  for (dyn = dyncontents; 
       dyn->d_tag != DT_NULL ; 
       dyn++) {
    switch (dyn->d_tag) {
    case DT_PLTGOT:
      pof->dt_pltgot = dyn->d_un.d_ptr;
      break;
    case DT_INIT:
      pof->dt_init = dyn->d_un.d_ptr;
      break;
    case DT_FINI:
      pof->dt_init = dyn->d_un.d_ptr;
      break;
    case DT_MIPS_TIME_STAMP:
      pof->dt_mips_time_stamp = dyn->d_un.d_val;
      break;
    case DT_MIPS_ICHECKSUM:
      pof->dt_mips_ichecksum = dyn->d_un.d_val;
      break;
    case DT_MIPS_LOCAL_GOTNO:
      tgotinfo[local_gotno_count++].local_gotno = dyn->d_un.d_val;
      break;
    case DT_MIPS_GOTSYM:
      tgotinfo[gotsym_count++].gotsym = dyn->d_un.d_val;
      break;
    }
  }
  free(dyncontents);
  pof->gotinfo = (gotinfo_t *)malloc(local_gotno_count * sizeof(gotinfo_t));
  for (i = 0; i < local_gotno_count; i++) {
    pof->gotinfo[i] = tgotinfo[i];
  }
  pof->ngots = local_gotno_count;
}

void 
load_liblist_section(process_ptr pprocess, object_file_t *pof, asection *section)
{
  bfd *abfd = pof->abfd;
  int size = bfd_section_size(abfd, section);
  Elf32_Lib *libs = (Elf32_Lib *)malloc(size);
  int nlibs = size/sizeof(Elf32_Lib);

  bfd_get_section_contents(abfd, section, libs, 0, size);
  // record each library for later loading
  cerr << "load_liblist_section: " << nlibs << " libs needed for file " << pof->filename << endl;

  pof->liblist = libs;
  pof->n_libs_needed = nlibs;
}

Elf32_Sym *
resolve_symbol(object_file_t *pof, Elf32_Sym *sym)
{
  int i;
  for (i = 0; i < pof->n_libs_needed; i++) {
    const char *name = 
      object_file_dynstr(pof, pof->liblist[i].l_name);
    char filename[1024];
    object_file_t *rpof;
    sprintf(filename, "/usr/shlib/%s", name);
    rpof = object_file_get(filename);
    {
      const char *name = object_file_dynstr(pof, sym->st_name);
      Elf32_Sym *defsym;
      defsym = object_file_dynsym_from_name(rpof, name);
      if (defsym && defsym->st_value != 0) {
	cerr << "resolve_symbox: found def for " << name << " in " << rpof->filename << endl;
	return defsym;
      }
    }
  }
  return NULL;
}

static void adjust_got_section(process_ptr pprocess, object_file_t *pof)
{
  /* object space address */
  ulong object_gotbase = pof->dt_pltgot;
  long nGot = pof->ngots;
  long i = 1;			/* skip over first NULL GOT entry */
  Elf32_Word n;

  
  cerr << "GOT: " << nGot << " GOT in " << pof->filename << endl;

  for (n = 0; n < nGot; n++) {
    gotinfo_t gotinfo = pof->gotinfo[n];

    cerr << 
      ".GOT[" << n << "]: \t {" << gotinfo.local_gotno << ", " << gotinfo.gotsym << "}" << endl;

    for (i = gotinfo.local_gotno; 1; i++) {
      Elf32_Addr *pgoti = 
	(Elf32_Addr *) pprocess->addr2phys(object_gotbase + i*sizeof(Elf32_Addr), NULL);
      Elf32_Addr goti = *pgoti;
      if (goti == 0 && i >= gotinfo.local_gotno) {
	cerr << "    end of GOT[" << n << "]" << endl
	     << "\t" << (void *) goti << " \t " << i-gotinfo.local_gotno << endl;

	// fprintf(stderr, "    end of GOT[%d]\n", n);
	// fprintf(stderr, "\t%12lx \t %ld\n", goti, i - gotinfo.local_gotno);
	break;
      }
      if (i == gotinfo.local_gotno) {
	cerr << "externals for GOT[" << n << "]:" << endl;
      }
      if (i < gotinfo.local_gotno) {
	cerr << "\t" << (void *) goti << "\t" << i - gotinfo.local_gotno << endl;
      } else {
	Elf32_Sym *sym = 
	  object_file_dynsym_from_index(pof, (i - gotinfo.local_gotno) + gotinfo.gotsym);

	cerr << "\t" << (void *) goti << " \t " << i-gotinfo.local_gotno
	     << " \t " << (void *) sym->st_value << " " << object_file_dynstr(pof, sym->st_name) << endl;

	if (1 || sym->st_value == 0) {
	  /* needs resolving */
	  Elf32_Sym *defsym = resolve_symbol(pof, sym);
	  if (defsym) {
	    cerr << "\t found definition: " << (void *) defsym->st_value << endl;
	    *pgoti = defsym->st_value;
	  }
	}
      }
    }
    i++;			/* skip over the NULL entry before starting the next GOT */
  }
}


/* 
 * Reads the text section of the object file and creates the linked list
 * of icode structures
 */

#if USE_ATOM_OUTPUT
static void
read_bb_info(process_ptr process, const char *Objname)
{

  icode_ptr picode;
  char bbname[80];
  char bbline[80];
  FILE *Fbb;
  long bbaddr;
  int bbinst, bbcode;

  bbline[0]=0;
  strcpy(bbname, Objname);
  Fbb = fopen(strcat(bbname, ".bb"), "rt");
  if (Fbb == NULL) {
    char command[2048];
    sprintf(command, "atom %s %s > %s",
	    Objname,
	    "/udir/pvega/AINT/src/bbdump.inst.o",
	    bbname);
    fprintf(Aint_output, "BBDUMP: %s ... ", command);
    system(command);
    fprintf(Aint_output, "done\n");
    Fbb = fopen(bbname, "rt");
    assert(Fbb != NULL);
  }

  while (fgets(bbline, 80, Fbb) != NULL) {
    sscanf(bbline, "%lx%x%d", &bbaddr, &bbinst, &bbcode);
    picode = addr2iphys(process, bbaddr, NULL);
    if (bbcode & BB_FIRST) 
      picode->iflags |= E_BB_FIRST;
    if (bbcode & BB_LAST) 
      picode->iflags |= E_BB_LAST;
  }
  fclose(Fbb);
}
#endif



/* 
 * analyze_code_segment(process, start, size):
 *    - look for conditional mov/store opportunities 
 */
static void analyze_code_segment(process_ptr pprocess, ulong start, ulong size)
{
#ifdef DO_ANALYZE_CS
  int i;
  informative("analyze_code_segment: startaddr=%lx, size=%ld",
	      start, size);
  process->n_instructions += size/sizeof(int);
  for (i = 0; i < size; i += sizeof(int)) {
    ulong textaddr = start + i;
    icode_ptr picode = addr2iphys(process, textaddr, NULL);
    if (picode->iflags & E_CONDBR) {
      if (picode->target->addr == textaddr + 8) {
	icode_ptr picode_op = picode->next;
	if (picode_op->iflags & (E_WRITE|E_READ)) {
	  fprintf(stderr, "    found conditional %s\n",
		  (picode_op->iflags & E_WRITE) ? "WRITE" : "READ");
	  fprintf(stderr, "    %16lx %8x \t %s $%d, %lx\n", 
		  textaddr, picode->instr, desc_table[picode->opnum].opname,
		  picode->args[RA], picode->target->addr);
	  fprintf(stderr, "    %16lx %8x \t %s $%d, %ld($%d)\n", 
		  textaddr+4, picode->instr, 
		  desc_table[picode_op->opnum].opname,
		  picode_op->args[RA], picode_op->immed, picode_op->args[RB]);
	} else if ((picode_op->iflags & E_BRANCH) == 0) {
	  int uses_lit = 
	    ((union alpha_instruction) ((unsigned)picode->instr)).o_format.form;
	  fprintf(stderr, "    found conditional OP\n");
	  fprintf(stderr, "    %16lx %8x \t %s $%d, %lx\n", 
		  textaddr, picode->instr, desc_table[picode->opnum].opname,
		  picode->args[RA], picode->target->addr);
	  fprintf(stderr, "    %16lx %8x \t %s $%d, %s%ld, $%d\n", 
		  textaddr+4, picode->instr, 
		  desc_table[picode_op->opnum].opname,
		  picode_op->args[RA], 
		  (uses_lit ? "" : "$"),
		  (uses_lit ? picode->literal : (ulong)picode_op->args[RB]),
		  picode_op->args[RC]);
	}
      }
    }
  }

#endif

}
