/* newinst.C
 * Assembly language routines for implementing instructions
 */


#include <iostream.h>
#include <stdio.h>
#include <sys/types.h>

#include <aint_types.h>
#include <aint.h>
#include "globals.h"
#include "opcodes.h"
#include "protos.h"

OP(unimplemented_op);

typedef unsigned char upixel8;
typedef unsigned short upixel16;
typedef signed char spixel8;
typedef signed short spixel16;

OP(perr_f)
{
  upixel8 *rav = (upixel8 *)&REG(RA);
  upixel8 *rbv = (upixel8 *)&REG(RB);
  upixel8 *rcv = (upixel8 *)&REG(RC);
  int i;
  for (i = 0; i < 8; i++) {
    if (rav[i] >= rbv[i]) {
      rcv[i] = rav[i] - rbv[i];
    } else {
      rcv[i] = rbv[i] - rav[i];
    }
  }
  return picode->next;
}

OP(minub8_f)
{
  upixel8 *rav = (upixel8 *)&REG(RA);
  upixel8 *rbv = (upixel8 *)&REG(RB);
  upixel8 *rcv = (upixel8 *)&REG(RC);
  int i;
  for (i = 0; i < 8; i++) {
    if (rav[i] >= rbv[i]) {
      rcv[i] = rbv[i];
    } else {
      rcv[i] = rav[i];
    }
  }
  return picode->next;
}

OP(minsb8_f)
{
  spixel8 *rav = (spixel8 *)&REG(RA);
  spixel8 *rbv = (spixel8 *)&REG(RB);
  spixel8 *rcv = (spixel8 *)&REG(RC);
  int i;
  for (i = 0; i < 8; i++) {
    if (rav[i] >= rbv[i]) {
      rcv[i] = rbv[i];
    } else {
      rcv[i] = rav[i];
    }
  }
  return picode->next;
}

OP(minuw4_f)
{
  upixel16 *rav = (upixel16 *)&REG(RA);
  upixel16 *rbv = (upixel16 *)&REG(RB);
  upixel16 *rcv = (upixel16 *)&REG(RC);
  int i;
  for (i = 0; i < 4; i++) {
    if (rav[i] >= rbv[i]) {
      rcv[i] = rbv[i];
    } else {
      rcv[i] = rav[i];
    }
  }
  return picode->next;
}

OP(minsw4_f)
{
  spixel16 *rav = (spixel16 *)&REG(RA);
  spixel16 *rbv = (spixel16 *)&REG(RB);
  spixel16 *rcv = (spixel16 *)&REG(RC);
  int i;
  for (i = 0; i < 4; i++) {
    if (rav[i] >= rbv[i]) {
      rcv[i] = rbv[i];
    } else {
      rcv[i] = rav[i];
    }
  }
  return picode->next;
}

OP(maxub8_f)
{
  upixel8 *rav = (upixel8 *)&REG(RA);
  upixel8 *rbv = (upixel8 *)&REG(RB);
  upixel8 *rcv = (upixel8 *)&REG(RC);
  int i;
  for (i = 0; i < 8; i++) {
    if (rav[i] <= rbv[i]) {
      rcv[i] = rbv[i];
    } else {
      rcv[i] = rav[i];
    }
  }
  return picode->next;
}

OP(maxsb8_f)
{
  spixel8 *rav = (spixel8 *)&REG(RA);
  spixel8 *rbv = (spixel8 *)&REG(RB);
  spixel8 *rcv = (spixel8 *)&REG(RC);
  int i;
  for (i = 0; i < 8; i++) {
    if (rav[i] <= rbv[i]) {
      rcv[i] = rbv[i];
    } else {
      rcv[i] = rav[i];
    }
  }
  return picode->next;
}

OP(maxuw4_f)
{
  upixel16 *rav = (upixel16 *)&REG(RA);
  upixel16 *rbv = (upixel16 *)&REG(RB);
  upixel16 *rcv = (upixel16 *)&REG(RC);
  int i;
  for (i = 0; i < 4; i++) {
    if (rav[i] <= rbv[i]) {
      rcv[i] = rbv[i];
    } else {
      rcv[i] = rav[i];
    }
  }
  return picode->next;
}

OP(maxsw4_f)
{
  spixel16 *rav = (spixel16 *)&REG(RA);
  spixel16 *rbv = (spixel16 *)&REG(RB);
  spixel16 *rcv = (spixel16 *)&REG(RC);
  int i;
  for (i = 0; i < 4; i++) {
    if (rav[i] <= rbv[i]) {
      rcv[i] = rbv[i];
    } else {
      rcv[i] = rav[i];
    }
  }
  return picode->next;
}

OP(pklb_f)
{
  upixel8 *rbv = (upixel8*)&REG(RB);
  upixel8 *rcv = (upixel8*)&REG(RC);

  REG(RC) = 0;
  rcv[0] = rbv[0];
  rcv[1] = rbv[4];

  return picode->next;
}

OP(pkwb_f)
{
  upixel8 *rbv = (upixel8*)&REG(RB);
  upixel8 *rcv = (upixel8*)&REG(RC);

  REG(RC) = 0;
  rcv[0] = rbv[0];
  rcv[1] = rbv[2];
  rcv[2] = rbv[4];
  rcv[3] = rbv[6];

  return picode->next;
}

OP(unpkbl_f)
{
  upixel8 *rbv = (upixel8*)&REG(RB);
  upixel8 *rcv = (upixel8*)&REG(RC);

  REG(RC) = 0;
  rcv[0] = rbv[0];
  rcv[4] = rbv[1];
  return picode->next;
}

OP(unpkbw_f)
{
  upixel8 *rbv = (upixel8*)&REG(RB);
  upixel8 *rcv = (upixel8*)&REG(RC);

  REG(RC) = 0;
  rcv[0] = rbv[0];
  rcv[2] = rbv[1];
  rcv[4] = rbv[2];
  rcv[6] = rbv[3];

  return picode->next;
}

OP(ctpop_f)
{
  int cnt = 0;
  int i;
  ulong rbv = REG(RB);
  for (i = 0; i < 64; i++) {
    if ((rbv >> i) & 1)
      cnt++;
  }
  REG(RC) = cnt;
  return picode->next;
}

OP(ctlz_f)
{
  int cnt = 0;
  int i;
  ulong rbv = REG(RB);
  for (i = 63; i >= 0; i--) {
    if ((rbv >> i) & 1)
      break;
    cnt++;
  }
  REG(RC) = cnt;
  return picode->next;
}

OP(cttz_f)
{
  int cnt = 0;
  int i;
  ulong rbv = REG(RB);
  for (i = 0; i < 64; i++) {
    if ((rbv >> i) & 1)
      break;
    cnt++;
  }
  REG(RC) = cnt;
  return picode->next;
}

OP(itofs_f)
{
  FP(RC) = (float)REG(RA);
  return picode->next;
}

UNIMPLEMENTED(itoff_f);

OP(itoft_f)
{
  FP(RC) = (double)REG(RA);
  return picode->next;
}

OP(ftois_f)
{
  float fav = (float)FP(RA);
  REG(RC) = *(int *)&fav;
  return picode->next;
}

OP(ftoit_f)
{
  REG(RC) = *(long *)&FP(RA);
  return picode->next;
}

OP(ecb_f)
{
  /* for now, does nothing in AINT */
  return picode->next;
}

OP(wh64_f)
{
  /* for now, does nothing in AINT */
  return picode->next;
}

