

// simple.C: a simple non-pipelined program-order processor simulation

#include <iostream.h>
#include <stdlib.h>
#include <string.h>

#include <aint.h>
// The catch-all include file that declares all relevant AINT types, consts,
// variables, and objects

const int SimpleThreads = 32;
int ThreadActive[SimpleThreads];
aint_addr_t PC[SimpleThreads];    // address to fetch from
int UsedThreads, CurrentThreads;

unsigned long InstCount;

// Aint calls this function before opening any object files - use this
// to initialize back-end variables
int
sim_init(int argc, char **argv, char **envp)
{
  /* AINT uses the getopt() function to parse the argument string, and
   * expects optind to have the correct value (first object file name)
   * upon return from this function. We must therefore use getopt() to
   * process arguments meant for us (delimited by "--").
   */

  extern int optind;    // AINT uses the getopt() function to keep track of args
  char c;
  int i;
  int dummy1=0;
  int dummy2=1;
  int dummy3=42;

  InstCount = 0;

  // some dummy args
  while ( (c=getopt(argc, argv, "f:d:i:")) != -1) {
    switch(c) {
    case 'f':
      dummy1=atoi(optarg);
      break;
    case 'd':
      dummy2=atoi(optarg);
      break;
    case 'i':
      dummy3=atoi(optarg);
      break;
    default:
      sim_usage(argc, argv);
      exit (1);
    }
  }

  cout << "Simple: running with dummy args " << dummy1 << ", " << dummy2 << " and " << dummy3 << endl;

  UsedThreads=0;
  CurrentThreads=0;
  for (i=0; i<SimpleThreads; i++) {
    ThreadActive[i] = 0;
  }
  // Return optind so that AINT finds the name of the target binary...
  return optind;
}


// Aint calls this function if the command line has an error
void
sim_usage(int argc, char **argv)
{
  cerr << endl << "Usage: <simname> [AINT options] executable" << endl;
  cerr << "This simulator takes no arguments, and accepts one executable" << endl;
}


void
sim_done()
{
  cout << endl << "Simple: Executed a total of " << InstCount << "Insts." << endl;
}

// AINT calls this function when it creates the main thread in the process
// corresponding to the target binary
void
sim_thread_begin(int tid, aint_addr_t pc)
{
  thread_ptr pthread = Threads[tid];    // Get thread pointer from AINT's table

  UsedThreads++;
  CurrentThreads++;
  ThreadActive[tid] = 1;

  AINT_create_path(tid, pc);    // Ask AINT to create main path

  informative("sim_thread_begin: began thread/path" << pthread->tid
      << " " << (void *) pthread->path_tree->pathid << endl);

  PC[tid] = pc;    // Starting address for thread
}

void
sim_thread_end(int tid)
{
  ThreadActive[tid] = 0;
  CurrentThreads--;
}

// The main execution loop - fetch, execute, commit each inst one at a time
void
sim_execute()
{
  int i;

  inflight_inst_ptr ifi;

  informative("sim_execute: commencing execution at " << (void *) PC[0] << endl);

  while (CurrentThreads) {
    for (i=0; i<UsedThreads; i++) {
      
      if (ThreadActive[i] == 0) continue;

      InstCount++;
      if (InstCount & ((1<<20)-1)); else cout << InstCount << " Insts" << endl;

      ifi = AINT_fetch(Threads[i]->path_tree, PC[i]);    // fetch next instruction
      informative("Fetch: " << (void *) PC[i] << endl);
      
      AINT_issue(ifi);    // ask AINT to issue the fetched instruction
      informative("Issue: " << (void *) PC[i] << endl);
      
      if (ifi->iflags & E_READ) {    // this is a load inst
	AINT_do_read(ifi);
      } else if (ifi->iflags & E_WRITE) {    // this is a store inst
	AINT_do_write(ifi);
      }
      
      PC[i] = ifi->nextPC;    // AINT updates ifi->nextPC upon executing inst
      if (ifi->iflags & E_WRITE) {
	informative("Committing 0x" << hex << ifi->get_pc() << ": " << hex << ifi->data << endl);
      } else {
	informative("Committing 0x" << hex << ifi->get_pc() << ": " << hex << Threads[ifi->tid()]->Reg[ifi->NewDest_Phys].Int64 << endl);
      }

      AINT_commit(ifi);    // Done
    }
  }

}  // sim_execute()

