/* simple.C
 * a non-pipelined, in-order processor simulator backend for MSAINT-multipath
 */

#include <iostream.h>
#include <aint.h>

#define MaxThreads 256

int *Tid;
aint_addr_t *NextPC;
ushort *ThreadActiveMask;
int ThreadsUsed;

void one_cycle(int tid, path_ptr ppath);

int
sim_init(int argc, char **argv, char **envp)
{
  Tid = new int[MaxThreads];
  NextPC = new aint_addr_t[MaxThreads];
  ThreadActiveMask = new ushort[MaxThreads];
  ThreadsUsed = 0;

  extern int optind;
  return optind;
}


void
sim_usage(int argc, char **argv)
{
}


void
sim_execute()
{
  // run until no threads alive
  int More;
  int tid;

  informative("sim_execute: commencing execution at " << (void *) NextPC[0] << endl);

  do {
    More=0;
    for (tid=0; tid < ThreadsUsed; tid++) {
      if (ThreadActiveMask[tid]) {
	More=1;
	one_cycle(tid, Threads[tid]->path_tree);
      }
    }
  } while(More);
  
}

void
one_cycle(int tid, path_ptr ppath)
{
  inflight_inst_ptr ifi;
  aint_addr_t PC = NextPC[tid];

  {
    ifi = AINT_fetch(ppath, PC);
    informative("Fetch: " << (void *) PC << endl);

    AINT_issue(ifi);
    informative("Issue: " << (void *) PC << endl);
    if (ifi->iflags & E_READ) {
      AINT_do_read(ifi);
    } else if (ifi->iflags & E_WRITE) {
      AINT_do_write(ifi);
    }

    PC = ifi->nextPC;

    if (ifi->iflags & E_WRITE) {
      informative("Committing " << (void *) ifi->picode->addr << " Wrote " 
		  << (void *) ifi->data << endl); 
    } else {
      informative("Committing " << (void *) ifi->picode->addr << " Dest " 
		  << (void *) ((ifi->picode->dest < MaxArgs)? ifi->iREG(ifi->picode->dest): 0x123456789abcdef0) << endl); 
    }
    AINT_commit(ifi);
  }
  NextPC[tid] = PC;
  if (PC == 0) ThreadActiveMask[tid] = 0;
}


void sim_thread_begin(int tid, aint_addr_t pc)
{
  thread_ptr pthread = Threads[tid];
  path_ptr ppath = AINT_create_path(tid, pc);

  ThreadActiveMask[tid] = 1;
  ThreadsUsed++;

  NextPC[tid] = pc;

  informative("sim_thread_begin: began thread/path" << pthread->tid
      << " " << (void *) pthread->path_tree->pathid << endl);
}

void sim_thread_end(int tid)
{
  ThreadActiveMask[tid] = 0;
}


void sim_thread_block(int tid)
{
  ThreadActiveMask[tid] = 0;
}

void sim_thread_unblock(int tid)
{
  ThreadActiveMask[tid] = 1;
}


void sim_done()
{
}

