#include <stdio.h>
#include <sys/types.h>
#include <math.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <errno.h>
/* #define N 256 */

int main(int argc, char *argv[])
{

  key_t k1, k2, k3;
  int shmid1, shmid2, shmid3;
  int pid, status, N;

  double *A, *B, *C;
  int seed, i0, i, j0, j, k;
  double sum;

  if (argc < 3) {
    printf("Usage: parmm size seed\n");
    exit(1);
  }
  sscanf(argv[1], "%d", &N);

  k1 = ftok("/vmunix", 42);
  k2 = ftok("/vmunix", 142);
  k3 = ftok("/vmunix", 242);

  shmid1 = shmget(k1, N*N*sizeof(double), IPC_CREAT|0666);
  if (shmid1 < 0) printf("Failure to get shmem seg 1, errno %d\n", errno);
  shmid2 = shmget(k2, N*N*sizeof(double), IPC_CREAT|0666);
  if (shmid2 < 0) printf("Failure to get shmem seg 2, errno %d\n", errno);
  shmid3 = shmget(k3, N*N*sizeof(double), IPC_CREAT|0666);
  if (shmid3 < 0) printf("Failure to get shmem seg 3, errno %d\n", errno);

  
  A = (double *) shmat(shmid1, 0, 0);
  B = (double *) shmat(shmid2, 0, 0);
  C = (double *) shmat(shmid3, 0, 0);


  /* Fill in the matrices A and B */
  sscanf(argv[2], "%d", &seed);
  srandom(seed);

  for (i=0; i<N; i++) for (j=0; j<N; j++) {
    int x1 = random();
    int x2 = random();
    int y1 = random();
    int y2 = random();
    A[i*N + j] = (x1 & 0xff) / (double) ((x2 & 0xff) + 1); 
    B[i*N + j] = (y1 & 0xff) / (double) ((y2 & 0xff) + 1);
  }

  /* fork off four processes. Important: Child must fork again */
  pid = 0;
  if (fork() == 0) { 
    pid ++;
    if (fork() == 0) {
      pid ++;
      if (fork() == 0) {
	pid ++;
      }
    }
  }

  if (pid) {
    A = (double *) shmat(shmid1, 0, 0);
    B = (double *) shmat(shmid2, 0, 0);
    C = (double *) shmat(shmid3, 0, 0);
  }

  /* Do matmul on our part */

  i0 = (pid%2)*(N/2);
  j0 = (pid/2)*(N/2);

  /*  printf("%d: i0 is %d j0 is %d\n", pid, i0, j0); */

  for (i=0; i<N/2; i++) for (j=0; j<N/2; j++) {
    int row, col;
    double t = 0;
    
    row = (i0 + i);
    col = (j0 + j);

#ifdef 0
    if (A[i*N + j] != i*N + j) printf("%d: Error: wrong value in A\n", pid); fflush(stdout);
    if (B[i*N + j] != i*N + j) printf("%d: Error: wrong value in B\n", pid); fflush(stdout);
#endif

    for (k=0; k<N; k++) t += A[row*N + k]*B[k*N + col];

    C[row*N + col] = t;
  }
  alpha_mb();

  wait(&status); 
  if (pid) exit(0);

  /* Last process - print out results */
  sum = 0.0;
  for (i=0; i<N; i++) for(j=0; j<N; j++) sum += C[i*N + j];

  printf("Sum is %g\n", sum);

  /* Remove the segments */
  shmctl(shmid1, IPC_RMID, NULL);
  shmctl(shmid2, IPC_RMID, NULL);
  shmctl(shmid3, IPC_RMID, NULL);

}
